use std::str::FromStr;

use color_eyre::eyre;
use eyre::eyre;

use crate::lib::{ParseLines, IterExt};


#[derive(Debug)]
enum Segment {
    A = 1,
    B = 2,
    C = 4,
    D = 8,
    E = 16,
    F = 32,
    G = 64, // wheee!
}
use Segment::*;

impl Segment {
    fn from_char(c: char) -> eyre::Result<Self> {
        match c {
            'a' => Ok(A),
            'b' => Ok(B),
            'c' => Ok(C),
            'd' => Ok(D),
            'e' => Ok(E),
            'f' => Ok(F),
            'g' => Ok(G),
            _ => Err(eyre!("Invalid character for segment: {}", c)),
        }
    }

    fn to_ord(&self) -> usize {
        // kinda gross but whatever
        match self {
            A => 0, B => 1, C => 2, D => 3,
            E => 4, F => 5, G => 6
        }
    }

    fn from_value(n: u8) -> eyre::Result<Self> {
        match n {
            1 => Ok(A),
            2 => Ok(B),
            4 => Ok(C),
            8 => Ok(D),
            16 => Ok(E),
            32 => Ok(F),
            64 => Ok(G),
            _ => Err(eyre!("Invalid value for segment: {}", n)),
        }
    }
}


#[derive(Copy, Clone, Default, Debug)]
struct SegmentSet {
    data: u8,
}

impl SegmentSet {
    fn new_filled() -> Self {
        SegmentSet {data: 255}
    }

    fn contains(&self, segment: Segment) -> bool {
        self.data & (segment as u8) > 0
    }

    fn insert(&mut self, segment: Segment) {
        self.data |= segment as u8;
    }

    fn remove(&mut self, segment: Segment) {
        self.data ^= segment as u8;
    }

    fn len(&self) -> u8 {
        (0..7).fold(0, |total, i| total + ((self.data >> i) & 1))
    }

    fn union(&self, other: &Self) -> Self {
        SegmentSet {data: self.data | other.data}
    }

    fn intersect(&self, other: &Self) -> Self {
        SegmentSet {data: self.data & other.data}
    }

    fn iter(&self) -> SegmentSetIter {
        SegmentSetIter {ss: self, pos: 0}
    }
}

impl FromStr for SegmentSet {
    type Err = eyre::Report;
    fn from_str(s: &str) -> eyre::Result<Self> {
        let mut set = Self::default();
        for c in s.chars() {
            set.insert(Segment::from_char(c)?);
        }
        Ok(set)
    }
}


struct SegmentSetIter<'a> {
    ss: &'a SegmentSet,
    pos: u8,
}

impl Iterator for SegmentSetIter<'_> {
    type Item = Segment;
    fn next(&mut self) -> Option<Segment> {
        while self.pos < 7 {
            let seg_value = self.ss.data & (1 << self.pos);
            self.pos += 1;
            if seg_value > 0 {
                return Some(Segment::from_value(seg_value).unwrap());
            }
        }
        return None
    }
}


#[derive(Default, Debug)]
struct Screen {
    patterns: [SegmentSet; 10],
    output: [SegmentSet; 4],
}

impl FromStr for Screen {
    type Err = eyre::Report;
    fn from_str(s: &str) -> eyre::Result<Self> {
        let mut screen = Screen::default();
        let (patterns, output) = s.split(" | ").take_pair()?;
        
        for (i, patt) in patterns.split(' ').enumerate() {
            screen.patterns[i] = patt.parse::<SegmentSet>()?;
        }
        for (i, out) in output.split(' ').enumerate() {
            screen.output[i] = out.parse::<SegmentSet>()?;
        }
        Ok(screen)
    }
}


// mapping of Segment -> SegmentSet, represents 
struct Candidates {
    signal: Segment,
    possible: SegmentSet
}


const DIGIT_SEGMENTS: [SegmentSet; 10] = [ // magic numbers, fun!
    SegmentSet {data: 0b01110111},
    SegmentSet {data: 0b00100100},
    SegmentSet {data: 0b01011101},
    SegmentSet {data: 0b01101101},
    SegmentSet {data: 0b00101110},
    SegmentSet {data: 0b01101011},
    SegmentSet {data: 0b01111011},
    SegmentSet {data: 0b00100101},
    SegmentSet {data: 0b01111111},
    SegmentSet {data: 0b01101111},
];


fn part1(screens: &[Screen]) -> usize {
    screens.iter()
        .flat_map(|d| d.output.iter())
        .filter(|s| matches!(s.len(), 2|3|4|7))
        .count()
}


pub fn run(data: &str) -> eyre::Result<(usize, usize)> {
    let screens = data
        .parse_lines::<Screen>()
        .collect::<Result<Vec<_>, _>>()?;



    Ok((part1(&screens), 0))
}
